// --- [ lagerfeuer.c - zur Kerze mutiert ] -----------------------------------
// ---   Kerze aus 2 led's, die durch die beiden PWM-Kanle von Counter0
//       unabhngig voneinander angesteuert werden.

#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <avr/wdt.h>

#include "parith-15.h"

// Port fr PWM
#define PORT_PWM 	PORTB  	//    
#define DDR_PWM  	DDRB   	// 
#define PIN_PWMA  0     	// OC0A
#define PIN_PWMB	1			// OC0B

#define IRQS_PER_SECOND  1000
#define PRESCALE 32

#define PWMSPEED1		160	// * 10us legt die Zeitabstnde zwischen 	
									// der kontinuierlichen nderungen des OC0A-Wertes
									// Geschwindigkeit der nderung
									// je kleiner, destro ruckhafter die nderung
#define PWMSPEED2		80
								
#define SPEED1			400 	// * 500 us die Zeit bis zur Neufestsetzung eines
									//  Zielwertes
									// je kleiner, desto hektischer
#define SPEED2			200

// 200 * 500us / 50 * 10us = 100.000 / 500 = 20 nderungen								
// Wird jede 0.5 Millisekunde um 1 verringert und bleibt bei 0 stehen
volatile uint16_t count_ms;

static uint16_t seed __attribute__ ((section (".noinit")));
void __init_seed(void) __attribute__ ((naked, section (".init8")));

//static void ioinit(void); 
void timers_init(void);

/*****************************************************************/
// Timer0: PWM  @ F_CPU/256, d.h. 31.25 kHz
// Timer1: IRQs @ 1kHz, d.h. alle 1ms
void timers_init (void)
{
 
	// TIMER1 gibt 1ms-Takt vor
	// Precale 16 -> 8.000.000 / (16 * 256) = 976  -> ca. 0.5ms
	TCCR1 = (0 << CTC1) | (1 << CS12) | (0 << CS11) | (1 << CS10); 
	// Timer1overflowenable
	TIMSK = (1 << TOIE1);
	
   // TIMER0 liefert das PWM-Signal
	// COM0A1 -> clear OC0A at compare match, set OC0A when downcounting
   // WGM00  -> phase correct, clear at top
	// Frequenz 8.000.000 / 510 * 1 = 15 KHz
   // TCCR0A = (1 << WGM00) | (1 << COM0A1);
	TCCR0A = ((1 << WGM00) | (1 << COM0A1 | 1<< COM0B1 | 0<<COM0B0));
   TCCR0B = (0 << CS01) | (1 << CS00); // Prescale = 1
}

/*****************************************************************/
// ISR wird alle 0.5 ms ausgefhrt
// Verringert `count_ms' um 1 falls noch nicht 0
//SIGNAL (SIG_OUTPUT_COMPARE1B)
//ISR(TIMER1_COMPA_vect)
ISR(TIMER1_OVF_vect)
{

if (count_ms) count_ms--;

}    

/*****************************************************************/
// Initialize hardware
/*
void ioinit (void) 
{
    // high fuse = 0xcc (BOD = 4.3V, WDT=on)
 //   wdt_enable (WDTO_15MS);
    
    // Alle (nicht verwendeten) Ports sind IN + PullUp
//    DDRB  = 0;
//    PORTB = 0x3f;
   
    // Enable Timer1 OutputCompare-IRQ
    // Set    Timer0 PWM
    timers_init();

     // OCRA = OUT + LOW
    DDR_PWM  |=  (1 << PIN_PWM);    
//    PORT_PWM &= ~(1 << BIT_PWM);    
} */

/*****************************************************************/
int main (void)
{
register uint8_t i = 0;
register uint8_t j;
register uint8_t pwmend;

// Initialisiere Hardware
timers_init();
DDRB = (1 << PIN_PWMA | 1<< PIN_PWMB); 
    
CLKPR = (1<<CLKPCE);	// Clock-division-faktor auf 1 setzen (default=8 -> 1MHz))
CLKPR = 0;
   
// Seed fr prandom() setzen
// seed wurde vor main() durch __init_seed() bestimmt
set_seed (seed);

sei(); 													// IRQs sind willkommen   
    
uint8_t ocra = 0, ocrb = 0;									// Zielwerte
uint8_t oa = 0, ob = 0;										// aktuelle Werte fr Compare

															// die Dauer der beiden Phasen 'zufllig' machen
uint8_t time1[] = { 40, 91, 38, 47, 61, 80, 57, 127};	// ruhige Zeiten
uint8_t time2[] = { 13,  9, 15,  7, 12, 17,  5,  17};	// Flackerzeiten

uint8_t limit1 = 0;									// von 0 bis litmit1 ruhige Zeiten
uint8_t limit2 = 0;									// von limit1 bis litmit2 Flackerzeiten
 
while (1)												// main loop
	{
// wdt_reset();
															// getrennte Counter fr beide PWM-Kanle
   if (oa < ocra) oa++;							 	// PWM wandert zum Sollwert ocr
   else if (oa > ocra) oa--;

   if (ob < ocrb) ob++;							 	// PWM wandert zum Sollwert ocr
   else if (ob > ocrb) ob--;

#define V 64  										// (v <= 128 !!) Maximale Helligkeit = (127 + v) <= 255			
															// bei leistungsstarken LEDS reichen kleinere Werte
															// je kleiner der WErt, desto grer der Kontrastumfang hell-dunkel
															
	OCR0A = (oa + V);									// compare setzen
	OCR0B = (ob + V); 								// in die obere Hlfte der Helligkeit schieben

   if (!count_ms)										// wenn counter_ms = 0
		{
         
      if (i == 0) 									// bei i == 0 neue Startwert auswhlen
			{
			ocra = (prandom() & 127);				// Bit.7 entfernen
			ocrb = (prandom() & 127);				// dadurch den Dynamikumfang reduzieren
			
			limit1 = time1[prandom() & 0x07];	// einen zuflligen Wert aus time1[] auswhlen
			limit2 = time2[prandom() & 0x07];	// einen zuflligen Wert aus time2[] auswhlen
			limit2 += limit1;
			}
	
	// in der ruhigen Phase wird mit geringer Geschwindigkeit der Helligkeitsnderung (PWMSPEED)
	// und relativen langen Phasen (SPEED) bis zum setzen neuer Zielwerte gearbeitet.
		
		if (i < limit1) count_ms = SPEED1; else count_ms = SPEED2;
		i++;        
		if (i < limit1)
			{
			ocra = (prandom() & 127);				// ruhigere Periode per Zufall bestimmen
			ocrb = (prandom() & 127);				// Dynamik reduzieren auf 7 Bit
			}
/*			
			{
			if (i & 0x01) ocr += 83; else ocr -= 57;  // oder durch Zhlen
			}*/
	
	// in der Flackerphase wird zwischen den Extremwerten gependelt
	// weiterhin wird Geschwindigkeit der Helligkeitsvernderung erhht 
		
		if (i > limit1 && i <= limit2)		// jetzt wird es turbulent
			{
			if (i & 0x01)
				{
				ocra = V;							// nur halber Dynamikbereich
				ocrb = V;							// damit die Kerze nicht dunkel wird
				}
				else 
				{
				ocra = 0;	
				ocrb = 0;
				}
			}
		if (i > limit2) i = 0;					// hier wird neu begonnen
		}	
		
	// hier wird eine Pause eingelegt, 
	// die die Geschwindigkeit der Helligkeitsvernderung regelt.
	if (i < (limit1 )) pwmend = PWMSPEED1; else pwmend = PWMSPEED2;
	j = 0;
		
	while (j < pwmend)							// unterschiedliche Zeiten anwenden
		{
		j++;
		_delay_us(10);								// etwas Ruhe in die Flamme bringen
		}
	}
}

// Bestimme seed aus zuflligem RAM-Inhalt
// !!! FUNKTION NICHT AUFRUFEN       !!!
// !!! FUNKTION WIRD AUTOMATISCH     !!!
// !!! IN .init8 vor main AUSGEFHRT !!!
void __init_seed (void)
{
	uint16_t s = 0;
	uint16_t *p = (uint16_t*) (RAMEND+1);
	extern uint16_t __noinit_start;

	while (p >= &__noinit_start + 1)
		s ^= * (--p);
        
	seed = s;
}

